from __future__ import absolute_import
import sys
import binascii

from .script import *
from .opcode import *
from .stack import Stack
from .standard import *

_bord = ord
if sys.version > '3':
    _bord = lambda x: x

half_order = 0x7fffffffffffffffffffffffffffffff5d576e7357a4501ddfe92f46681b20a0

class Engine(object):
    """Virtual machine that executes scripts."""
    def __init__(self, script_pub_key, tx, tx_idx, flags, script_version):
        self.script_idx = 0
        self.script_off = 0
        self.last_code_sep = 0
        self.dstack = Stack()
        self.astack = Stack()
        self.cond_stack = []
        self.num_ops = 0
        self.bip16 = False
        self.saved_first_stack = []

        if tx_idx < 0 or tx_idx >= len(tx.txins):
            raise InvalidIndexError()
        script_sig = tx.txins[tx_idx].sig_script

        # The clean stack flag (ScriptVerifyCleanStack) is not allowed without
        # the pay-to-script-hash (P2SH) evaluation (ScriptBip16) flag.
        #
        # Recall that evaluating a P2SH script without the flag set results in
        # non-P2SH evaluation which leaves the P2SH inputs on the stack.  Thus,
        # allowing the clean stack flag without the P2SH flag would make it
        # possible to have a situation where P2SH would not be a soft fork when
        # it should be.
        self.version = script_version
        self.flags = flags
        if self.has_flag(ScriptVerifyCleanStack) and not self.has_flag(ScriptBip16):
            raise InvalidFlagsError()

        # The signature script must only contain data pushes is the associated flag is set.
        if self.has_flag(ScriptVerifySigPushOnly) and not IsPushOnlyScript(script_sig):
            raise StackNonPushOnlyError()

        # Subscripts for P2SH outputs are not allowed to use any stake tag opcodes
        # if script version is 0.
        if script_version == DefaultScriptVersion:
            HasP2SHScriptSigStakeOpCodes(script_version, script_sig, script_pub_key)

        # The engine stores the scripts in parsed form.
        # This allows multiple scripts to be executed in sequence.
        scripts = [script_sig, script_pub_key]
        self.scripts = []
        for scr in scripts:
            if len(scr) > maxScriptSize:
                raise StackLongScriptError()
            self.scripts.append(parseScript(scr))

        # Advance the program counter to the public key script if the signature
        # script is empty since there is nothing to execute for it in that case.
        if len(scripts[0]) == 0:
            self.script_idx += 1

        if self.has_flag(ScriptBip16) and isAnyKindOfScriptHash(self.scripts[1]):
            if not isPushOnly(self.scripts[0]):
                raise StackP2SHNonPushOnlyError()
            self.bip16 = True
        if self.has_flag(ScriptVerifyMinimalData):
            self.dstack.verify_minimal_data = True
            self.astack.verify_minimal_data = True

        self.tx = tx
        self.tx_idx = tx_idx

    def has_flag(self, flag):
        return self.flags & flag == flag

    def is_branch_executing(self):
        """Returns whether the current conditional branch is actively executing.

        For example, when the data stack has an OP_FALSE on it and an
        OP_IF is encountered, the branch is inactive until an OP_ELSE or
        OP_ENDIF is encountered.
        """
        if len(self.cond_stack) == 0:
            return True
        return self.cond_stack[-1] == OpCondTrue

    def execute_opcode(self, pop):
        if pop.isDisabled():
            raise StackOpDisabledError()

        if pop.alwaysIllegal():
            raise StackReservedOpcodeError()

        # Note that this includes OP_RESERVED which counts as a push operation.
        if pop.opcode.value > OP_16:
            self.num_ops += 1
            if self.num_ops > MaxOpsPerScript:
                raise StackTooManyOperationsError()
        elif len(pop.data) > MaxScriptElementSize:
            raise StackElementTooBigError()

        # Nothing left to do when this is not a conditional opcode and it's not
        # in an executing branch.
        if not self.is_branch_executing() and not pop.isConditional():
            return

        # Ensure all executed data push opcods use minimal encoding if set.
        if self.dstack.verify_minimal_data and self.is_branch_executing() and \
                    pop.opcode.value >= 0 and pop.opcode.value <= OP_PUSHDATA4:
            pop.checkMinimalDataPush()

        return pop.opcode.opfunc(pop, self)

    def disasm(self, scriptIdx, scriptOff, verbose=True):
        """Helper function to produce the output for disasm_pc and disasm_script."""
        if verbose:
            return '%02x:%04x: %s' % (scriptIdx, scriptOff, self.scripts[scriptIdx][scriptOff].human_readable(oneline=False))
        return '%s' % self.scripts[scriptIdx][scriptOff].human_readable(oneline=False)

    def valid_pc(self):
        """Raises an exception if the current script position is valid for execution."""
        if self.script_idx >= len(self.scripts):
            raise Exception('past input scripts %s:%s %s:xxxx' % (self.script_idx, self.script_off, len(self.scripts)))

    def cur_pc(self):
        """Returns either the current script and offset or raises an error if the position isn't valid."""
        self.valid_pc()
        return self.script_idx, self.script_off

    def disasm_pc(self, verbose=True):
        """Returns the string for the disassembly of the opcode that will execute next when step() is called."""
        scriptIdx, scriptOff = self.cur_pc()
        return self.disasm(scriptIdx, scriptOff, verbose)

    def disasm_script(self, idx, verbose=True):
        """Returns the disassembly string for the script at offset idx.

        Index 0 is the signature script and 1 is the public key script.
        """
        if idx >= len(self.scripts):
            raise StackInvalidIndexError()

        disstr = ''
        for i in range(len(self.scripts[idx])):
            disstr += self.disasm(idx, i, verbose) + '\n'
        return disstr

    def check_error_condition(self, final_script):
        """Raises an exception if the script has not finished."""
        # Check execution is actually done.
        if self.script_idx < len(self.scripts):
            raise StackScriptUnfinishedError()
        if final_script and self.has_flag(ScriptVerifyCleanStack) and self.dstack.depth() != 1:
            raise StackCleanStackError()
        elif self.dstack.depth() < 1:
            raise StackEmptyStackError()

        v = self.dstack.peek_bool(0)
        if not v:
            raise StackScriptFailedError()

    def step(self):
        """Executes the next instruction and moves the program counter to the next opcode.

        Moves to the next script is the current one has ended. Step will return True
        if the last opcode was successfully executed.
        """
        self.valid_pc()
        opcode = self.scripts[self.script_idx][self.script_off]

        # Execute the opcode while taking into account several things such as
        # disabled opcodes, illegal opcodes, max allowed operations per script,
        # max script element sizes, and conditionals.
        self.execute_opcode(opcode)

        # The number of elements in the combination of the data and alt stacks must
        # not exceed the maximum number of stack elements allowed.
        if self.dstack.depth() + self.astack.depth() > maxStackSize:
            raise StackOverflowError()

        # Prepare for the next instruction.
        self.script_off += 1
        if self.script_off >= len(self.scripts[self.script_idx]):
            # Illegal to have an if that straddles two scripts.
            if len(self.cond_stack) != 0:
                raise StackMissingEndifError()

            # Alt stack doesn't persist.
            if self.astack.depth():
                self.astack.drop(self.astack.depth())

            self.num_ops = 0
            self.script_off = 0
            if self.script_idx == 0 and self.bip16:
                self.script_idx += 1
                self.saved_first_stack = self.get_stack()
            elif self.script_idx == 1 and self.bip16:
                # Put us past the end for check_error_condition()
                self.script_idx += 1
                # Check script ran successfully and pull the script
                # out of the first stack and execute that.
                self.check_error_condition(False)

                script = self.saved_first_stack[-1]
                pops = parseScript(script)
                self.scripts.append(pops)

                # Set stack to be the stack from the first script minus the script itself.
                self.set_stack(self.saved_first_stack[:-1])
            else:
                self.script_idx += 1

            # There are zero-length scripts in the wild.
            if (self.script_idx < len(self.scripts) and
                    self.script_off >= len(self.scripts[self.script_idx])):
                self.script_idx += 1

            self.last_code_sep = 0
            if self.script_idx >= len(self.scripts):
                return True

        return False

    def execute(self):
        """Execute all scripts in the engine and raise an exception if one occurs."""
        if self.version != DefaultScriptVersion:
            return

        done = False
        while done != True:
            done = self.step()

        return self.check_error_condition(True)

    def sub_script(self):
        """Returns the script since the last OP_CODESEPARATOR."""
        return self.scripts[self.script_idx][self.last_code_sep:]

    def check_hash_type_encoding(self, hash_type):
        """Raises an exception if the passed hashtype doesn't adhere to the strict encoding requirements (if enabled)."""
        if not self.has_flag(ScriptVerifyStrictEncoding):
            return

        sigHashType = hash_type & ~SigHashAnyOneCanPay
        if sigHashType < SigHashAll or sigHashType > SigHashSingle:
            raise Exception('invalid hashtype: 0x%x' % hash_type)

    def check_pubkey_encoding(self, pub_key):
        """Raises an exception if the passed public key doesn't adhere to the strict encoding requirements (if enabled)."""
        if not self.has_flag(ScriptVerifyStrictEncoding):
            return

        # Compressed public key.
        if len(pub_key) == 33 and (_bord(pub_key[0]) == 0x02 or _bord(pub_key[0]) == 0x03):
            return
        # Uncompressed public key.
        if len(pub_key) == 65 and _bord(pub_key[0]) == 0x04:
            return
        raise StackInvalidPubKeyError()

    def check_signature_encoding(self, sig):
        """Raises an exception if the passed signature doesn't adhere to the strict encoding requirements (if enabled)."""
        if (not self.has_flag(ScriptVerifyDERSignatures) and
            not self.has_flag(ScriptVerifyLowS) and
            not self.has_flag(ScriptVerifyStrictEncoding)):
            return

        # The format of a DER encoded signature is as follows:
        #
        # 0x30 <total length> 0x02 <length of R> <R> 0x02 <length of S> <S>
        #   - 0x30 is the ASN.1 identifier for a sequence
        #   - Total length is 1 byte and specifies length of all remaining data
        #   - 0x02 is the ASN.1 identifier that specifies an integer follows
        #   - Length of R is 1 byte and specifies how many bytes R occupies
        #   - R is the arbitrary length big-endian encoded number which
        #     represents the R value of the signature.  DER encoding dictates
        #     that the value must be encoded using the minimum possible number
        #     of bytes.  This implies the first byte can only be null if the
        #     highest bit of the next byte is set in order to prevent it from
        #     being interpreted as a negative number.
        #   - 0x02 is once again the ASN.1 integer identifier
        #   - Length of S is 1 byte and specifies how many bytes S occupies
        #   - S is the arbitrary length big-endian encoded number which
        #     represents the S value of the signature.  The encoding rules are
        #     identical as those for R.

        # Minimum length is when both numbers are 1 byte each.
        # 0x30 + <1-byte> + 0x02 + 0x01 + <byte> + 0x2 + 0x01 + <byte>
        if len(sig) < 8:
            raise Exception('malformed signature: too short: %d < 8' % len(sig))

        # Maximum length is when both numbers are 33 bytes each.  It is 33
        # bytes because a 256-bit integer requires 32 bytes and an additional
        # leading null byte might required if the high bit is set in the value.
        # 0x30 + <1-byte> + 0x02 + 0x21 + <33 bytes> + 0x2 + 0x21 + <33 bytes>
        if len(sig) > 72:
            raise Exception('malformed signature: too long: %d > 72' % len(sig))

        if _bord(sig[0]) != 0x30:
            raise Exception('malformed signature: format has wrong type: 0x%x' % sig[0])

        if _bord(sig[1]) != len(sig) - 2:
            raise Exception('malformed signature: bad length: %d != %d' % (sig[1], len(sig)-2))

        rlen = _bord(sig[3])

        # Make sure S is inside the signature.
        if rlen + 5 > len(sig):
            raise Exception('malformed signature: S out of bounds')

        slen = _bord(sig[rlen+5])

        # The length of the elements does not match the length of the signature.
        if rlen + slen + 6 != len(sig):
            raise Exception('malformed signature: invalid R length')

        # R elements must be integers.
        if _bord(sig[2]) != 0x02:
            raise Exception('malformed signature: missing first integer marker')

        # Zero-length integers are not allowed for R.
        if rlen == 0:
            raise Exception('malformed signature: R length is zero')

        # R must not be negative
        if _bord(sig[4]) & 0x80 != 0:
            raise Exception('malformed signature: R value is negative')

        # Null bytes at the start of R are not allowed unless R would
        # otherwise be interpreted as a negative number.
        if rlen > 1 and _bord(sig[4]) == 0x00 and _bord(sig[5]) & 0x80 == 0:
            raise Exception('malformed signature: invalid R value')

        # S elements must be integers.
        if _bord(sig[rlen+4]) != 0x02:
            raise Exception('malformed signature: missing second integer marker')

        # Zero-length integers are not allowed for S.
        if slen == 0:
            raise Exception('malformed signature: S length is zero')

        # S must not be negative.
        if _bord(sig[rlen+6]) & 0x80 != 0:
            raise Exception('malformed signature: S value is negative')

        # Null bytes at the start of S are not allowed unless S would
        # otherwise be interpreted as a negative number.
        if slen > 1 and _bord(sig[rlen+6]) == 0x00 and _bord(sig[rlen+7]) & 0x80 == 0:
            raise Exception('malformed signature: invalid S value')

        # Verify the S value is <= half the order of the curve.  This check is
        # done because when it is higher, the complement modulo the order can
        # be used instead which is a shorter encoding by 1 byte.  Further,
        # without enforcing this, it is possible to replace a signature in a
        # valid transaction with the complement while still being a valid
        # signature that verifies.  This would result in changing the
        # transaction hash and thus is source of malleability.
        if self.has_flag(ScriptVerifyLowS):
            svalue = int(binascii.hexlify(sig[rlen+6 : rlen+6+slen]), 16)
            if svalue > half_order:
                raise StackInvalidLowSSignatureError()
            pass

    def get_stack(self):
        """Returns the contents of the primary stack."""
        return get_stack(self.dstack)

    def set_stack(self, data):
        """Sets the contents of the primary stack."""
        set_stack(self.dstack, data)

    def get_alt_stack(self):
        """Returns the contents of the alt stack."""
        return get_stack(self.astack)

    def set_alt_stack(self, data):
        """Sets the contents of the alt stack."""
        set_stack(self.astack, data)

def get_stack(stack):
    arr = []
    for i in range(stack.depth()):
        arr.append(stack.peek_bytearray(i))
    arr.reverse()
    return b''.join(arr)

def set_stack(stack, data):
    stack.drop(stack.depth())
    for i in data:
        stack.push_bytearray(i)



"""Script-related errors."""

# Engine execution errors.

class EngineExecutionError(Exception):
    pass

class StackShortScriptError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'execute past end of script')

class StackLongScriptError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'script is longer than maximum allowed')

class StackUnderflowError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'stack underflow')

class StackInvalidArgsError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'invalid argument')

class StackOpDisabledError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Disabled Opcode')

class StackVerifyFailedError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Verify failed')

class StackNumberTooBigError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'number too big')

class StackInvalidOpcodeError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Invalid Opcode')

class StackReservedOpcodeError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Reserved Opcode')

class StackEarlyReturnError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Script returned early')

class StackNoIfError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'OP_ELSE or OP_ENDIF with no matching OP_IF')

class StackMissingEndifError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'execute fail, in conditional execution')

class StackTooManyPubkeysError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Invalid pubkey count in OP_CHECKMULTISIG')

class StackTooManyOperationsError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Too many operations in script')

class StackElementTooBigError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Element in script too large')

class StackUnknownAddressError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'non-recognised address')

class StackScriptFailedError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'execute fail, fail on stack')

class StackScriptUnfinishedError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Error check when script unfinished')

class StackEmptyStackError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Stack empty at end of execution')

class StackP2SHNonPushOnlyError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'pay to script hash with non-pushonly input')

class StackInvalidParseTypeError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'internal error: invalid parsetype found')

class StackInvalidAddrOffsetError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'internal error: invalid offset found')

class StackInvalidIndexError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Invalid script index')

class StackNonPushOnlyError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'SigScript is non-pushonly')

class StackOverflowError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'Stacks overflowed')

class StackInvalidLowSSignatureError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'invalid low s signature')

class StackInvalidPubKeyError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'invalid strict pubkey')

class StackCleanStackError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'stack is not clean')

class StackMinimalDataError(EngineExecutionError):
    def __init__(self, *args, **kwargs):
        EngineExecutionError.__init__(self, 'non-minimally encoded script number')


# Engine script errors.
class EngineScriptError(Exception):
    pass

class InvalidFlagsError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'invalid flags combination')

class InvalidIndexError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'invalid input index')

class UnsupportedAddressError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'unsupported address type')

class BadNumRequiredError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'more signatures required than keys present')

class SighashSingleIdxError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'invalid SIGHASH_SINGLE script index')

class SubstrIndexNegativeError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'negative number given for substring index')

class SubstrIdxOutOfBoundsError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'out of bounds number given for substring index')

class NegativeRotationError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'rotation depth negative')

class RotationOverflowError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'rotation depth out of bounds')

class NegativeShiftError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'shift depth negative')

class ShiftOverflowError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'shift depth out of bounds')

class DivideByZeroError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'division by zero')

class P2SHStakeOpCodesError(EngineScriptError):
    def __init__(self, *args, **kwargs):
        EngineScriptError.__init__(self, 'stake opcodes were found in a p2sh script')




from __future__ import absolute_import

from bitcoin.core.scripteval import _CastToBool

from .scriptnum import num2bytearray, script_num, math_op_code_max_script_num_len
from .errors import *

class Stack(object):
    def __init__(self, stack=None, verify_minimal_data=False):
        if stack is None:
            stack = []
        self.stack = stack
        self.verify_minimal_data = verify_minimal_data

    def __str__(self):
        return ' '.join([i.encode('hex') for i in self.stack])

    def depth(self):
        return len(self.stack)

    def push_bytearray(self, val):
        self.stack.append(val)

    def push_int(self, val):
        self.push_bytearray(num2bytearray(val))

    def push_bool(self, val):
        self.push_bytearray(b'\x01' if val else b'\x00')

    def pop_bytearray(self):
        return self.nip(0)

    def pop_int(self, max_len):
        val = self.pop_bytearray()
        return script_num(val, self.verify_minimal_data, max_len)

    def pop_bool(self):
        val = self.pop_bytearray()
        return _CastToBool(val)

    def peek_bytearray(self, idx):
        size = len(self.stack)
        if idx < 0 or idx >= size:
            raise StackUnderflowError()
        return self.stack[size - idx - 1]

    def peek_int(self, idx):
        val = self.peek_bytearray(idx)
        return script_num(val, self.verify_minimal_data, math_op_code_max_script_num_len)

    def peek_bool(self, idx):
        val = self.peek_bytearray(idx)
        return _CastToBool(val)

    def nip(self, idx):
        """Remove the idx-th stack item."""
        size = len(self.stack)
        if idx < 0 or idx > size - 1:
            raise StackUnderflowError()
        return self.stack.pop(size - idx - 1)

    def tuck(self):
        """Copy the top stack item to the second-to-top stack index."""
        val2 = self.pop_bytearray()
        val1 = self.pop_bytearray()
        self.push_bytearray(val2)
        self.push_bytearray(val1)
        self.push_bytearray(val2)

    def drop(self, n):
        """Remove the top n items from the stack."""
        if n < 1:
            raise StackInvalidArgsError()
        for i in range(n):
            self.pop_bytearray()

    def dup(self, n):
        """Duplicate the top n items on the stack."""
        if n < 1:
            raise StackInvalidArgsError()
        for i in range(n, 0, -1):
            val = self.peek_bytearray(n - 1)
            self.push_bytearray(val)

    def rot(self, n):
        """Rotate the top 3n items on the stack to the left n times."""
        if n < 1:
            raise StackInvalidArgsError()
        entry = 3*n - 1
        for i in range(n, 0, -1):
            val = self.nip(entry)
            self.push_bytearray(val)

    def swap(self, n):
        """Swap the top n items on the stack with those below them."""
        if n < 1:
            raise StackInvalidArgsError()
        entry = 2*n - 1
        for i in range(n, 0, -1):
            val = self.nip(entry)
            self.push_bytearray(val)

    def over(self, n):
        """Copy n items n items back to the top of the stack."""
        if n < 1:
            raise StackInvalidArgsError()
        entry = 2*n - 1
        while n > 0:
            val = self.peek_bytearray(entry)
            self.push_bytearray(val)
            n -= 1

    def pick(self, n):
        """Copy the item n items back in the stack to the top."""
        val = self.peek_bytearray(n)
        self.push_bytearray(val)

    def roll(self, n):
        """Move the item n items back in the stack to the top."""
        val = self.nip(n)
        self.push_bytearray(val)





# These are here to avoid circular imports.


# Script Flags #

# ScriptBip16 defines whether the bip16 threshhold has passed and thus
# pay-to-script hash transactions will be fully validated.
ScriptBip16 = 1

# ScriptStrictMultiSig defines whether to verify the stack item
# used by CHECKMULTISIG is zero length.
ScriptStrictMultiSig = 1 << 1

# ScriptDiscourageUpgradableNops defines whether to verify that
# NOP1 through NOP10 are reserved for future soft-fork upgrades.  This
# flag must not be used for consensus critical code nor applied to
# blocks as this flag is only for stricter standard transaction
# checks.  This flag is only applied when the above opcodes are
# executed.
ScriptDiscourageUpgradableNops = 1 << 2

# ScriptVerifyCheckLockTimeVerify defines whether to verify that
# a transaction output is spendable based on the locktime.
# This is BIP0065.
ScriptVerifyCheckLockTimeVerify = 1 << 3

# ScriptVerifyCleanStack defines that the stack must contain only
# one stack element after evaluation and that the element must be
# true if interpreted as a boolean.  This is rule 6 of BIP0062.
# This flag should never be used without the ScriptBip16 flag.
ScriptVerifyCleanStack = 1 << 4

# ScriptVerifyDERSignatures defines that signatures are required
# to compily with the DER format.
ScriptVerifyDERSignatures = 1 << 5

# ScriptVerifyLowS defines that signtures are required to comply with
# the DER format and whose S value is <= order / 2.  This is rule 5
# of BIP0062.
ScriptVerifyLowS = 1 << 6

# ScriptVerifyMinimalData defines that signatures must use the smallest
# push operator. This is both rules 3 and 4 of BIP0062.
ScriptVerifyMinimalData = 1 << 7

# ScriptVerifySigPushOnly defines that signature scripts must contain
# only pushed data.  This is rule 2 of BIP0062.
ScriptVerifySigPushOnly = 1 << 8

# ScriptVerifyStrictEncoding defines that signature scripts and
# public keys must follow the strict encoding requirements.
ScriptVerifyStrictEncoding = 1 << 9

maxStackSize = 1024
maxScriptSize = 16384
DefaultScriptVersion = 0


SigHashOld = 0x0
SigHashAll = 0x1
SigHashNone = 0x2
SigHashSingle = 0x3
SigHashAllValue = 0x4
SigHashAnyOneCanPay = 0x80
sigHashMask = 0x1f

MaxOpsPerScript = 255
MaxPubKeysPerMultiSig = 20
MaxScriptElementSize = 2048




from __future__ import absolute_import
from io import BytesIO
import struct
import sys

from decred.core.chaincfg import sig_hash_optimization
from decred.core.serialize import DecredHash
from .opcode import *
from .constants import *

_bord = ord
if sys.version > '3':
    _bord = lambda x: x

def isSmallInt(op):
    """Returns whether or not op is considered a small integer."""
    if op.value == OP_0 or (op.value >= OP_1 and op.value <= OP_16):
        return True
    return False

def IsPayToScriptHash(script):
    """Returns whether script is in the standard P2SH format."""
    pops = parseScript(script)
    return isScriptHash(pops)

def isPushOnly(pops):
    """Returns whether the script only pushes data."""
    for pop in pops:
        if pop.opcode.value > OP_16:
            return False
    return True

def IsPushOnlyScript(script):
    pops = parseScript(script)
    return isPushOnly(pops)

#def HasP2SHScriptSigStakeOpCodes(version, scriptSig, scriptPubKey):
# Moved to standard.py

def parseScriptTemplate(script, opcodes):
    retScript = []
    i = 0
    while i < len(script):
        instr = _bord(script[i])
        op = opcodes[instr]
        pop = ParsedOpcode(op, b'')
        # Parse data out of instruction.

        # No additional data.
        if op.length == 1:
            i += 1
        # Data pushes of specific length (e.g. OP_DATA_20)
        elif op.length > 1:
            if len(script[i:]) < op.length:
                raise StackShortScriptError()
            pop.data = script[i+1 : i+op.length]
            i += op.length
        # Data pushes with parsed lengths (e.g. OP_PUDHDATAP1)
        elif op.length < 0:
            L = 0
            off = i + 1
            if len(script[off:]) < -op.length:
                raise StackShortScriptError()

            # Next -length bytes are little endian length of data.
            if op.length == -1:
                L = script[off]
            elif op.length == -2:
                L = (script[off+1] << 8) | script[off]
            elif op.length == -4:
                L = (script[off+3] << 24) | (script[off+2] << 16) | (script[off+1] << 8) | script[off]
            else:
                raise Exception('invalid opcode length: %d' % op.length)

            # Move offset to beginning of the data.
            off += -op.length

            # Disallow entries that do not fit script or were sign extended.
            if L > len(script[off:]) or L < 0:
                raise StackShortScriptError()

            pop.data = script[off: off+L]
            i += 1 - op.length + L

        retScript.append(pop)

    return retScript

def ParseScript(script):
    return parseScriptTemplate(script, opcodeArray)

def parseScript(script):
    """Preparses the script in bytes into a list of ParsedOpcodes while applying sanity checks."""
    return parseScriptTemplate(script, opcodeArray)

def unparseScript(pops):
    """Reverses the action of parseScript."""
    script = b''
    for pop in pops:
        b = pop.as_bytes()
        script += b
    return script

def DisasmString(buf):
    """Formats the disassembled script for one line printing."""
    disbuf = ''
    opcodes = parseScript(buf)
    for pop in opcodes:
        disbuf += pop.human_readable(oneline=True)
    if disbuf:
        disbuf = disbuf[:-1]
    return disbuf

def removeOpcode(pkscript, opcode):
    """Removes any occurrence of opcode from pkscript."""
    retScript = b''
    for pop in pkscript:
        if pop.opcode.value != opcode:
            retScript += pop
    return retScript

def canonicalPush(pop):
    """Returns True if either not a push instruction or the push is canonical."""
    op = pop.opcode.value
    data = pop.data
    dlen = len(pop.data)

    if op > OP_16:
        return True

    if op < OP_PUSHDATA1 and op > OP_0 and (dlen == 1 and data[0] <= 16):
        return False
    if op == OP_PUSHDATA1 and dlen < OP_PUSHDATA1:
        return False
    if op == OP_PUSHDATA2 and dlen <= 0xff:
        return False
    if op == OP_PUSHDATA4 and dlen <= 0xffff:
        return False
    return True

def removeOpcodeByData(pkscript, data):
    """Returns the script minus any opcodes that would push data."""
    retScript = b''
    for pop in pkscript:
        if not canonicalPush(pop) or not data in pop.data:
            retScript += pop
    return retScript

def CalcSignatureHash(script, hashType, tx, idx, cachedPrefix):
    return calcSignatureHash(script, hashType, tx, idx, cachedPrefix)

def calcSignatureHash(script, hashType, tx, idx, cachedPrefix):
    """Calculate the signature hash to be used for signing and verification."""
    pass

    # The SigHashSingle signature type signs only the corresponding input
    # and output (the output with the same index number as the input).
    #
    # Since transactions can have more inputs than outputs, this means it
    # is improper to use SigHashSingle on input indices that don't have a
    # corresponding output.
    #
    # A bug in the original Satoshi client implementation means specifying
    # an index that is out of range results in a signature hash of 1 (as a
    # uint256 little endian).  The original intent appeared to be to
    # indicate failure, but unfortunately, it was never checked and thus is
    # treated as the actual signature hash.  This buggy behavior is now
    # part of the consensus and a hard fork would be required to fix it.
    #
    # Due to this, care must be taken by software that creates transactions
    # which make use of SigHashSingle because it can lead to an extremely
    # dangerous situation where the invalid inputs will end up signing a
    # hash of 1.  This in turn presents an opportunity for attackers to
    # cleverly construct transactions which can steal those coins provided
    # they can reuse signatures.
    #
    # Decred mitigates this by actually returning an error instead.
    if hashType&sigHashMask == SigHashSingle and idx >= len(tx.TxOut):
            raise SighashSingleIdxError()

    # Remove all instances of OP_CODESEPARATOR from the script.
    script = removeOpcode(script, OP_CODESEPARATOR)

    # Make a deep copy of the transaction, zeroing out the script for all
    # inputs that are not currently being processed.
    txCopy = Transaction.from_tx(tx)
    for i in range(len(txCopy.txins)):
        txIn = txCopy.txins[i]
        if i == idx:
            sigScript = unparseScript(script)
            txCopy.txins[idx].sig_script = sigScript
        else:
            txCopy.txins[i].sig_script = b''

    # Default behavior has all outputs set up.
    # for i := range txCopy.TxOut {
    #         var txOut wire.TxOut
    #         txOut = *txCopy.TxOut[i]
    #         txCopy.TxOut[i] = &txOut
    # }

    htype = hashType & sigHashMask
    if htype == SigHashNone:
        txCopy.txouts = txCopy.txouts[0:0] # Empty slice.
        for i, txin in enumerate(txCopy.txins):
            if i != idx:
                txin.sequence = 0
    elif htype == SigHashSingle:
        # Resize outputs to up to and including requested index.
        txCopy.txouts = txCopy.txouts[:idx+1]

        # All but the current output gets zeroed out.
        for i, txout in enumerate(txCopy.txouts):
            if i < idx:
                txout.value = -1
                txout.pk_script = b''

        # Sequence on all other inputs is 0, too.
        for i, txin in enumerate(txCopy.txins):
            if i != idx:
                txin.sequence = 0

    if hashType & SigHashAnyOneCanPay != 0:
        txCopy.txins = txCopy.txins[idx : idx+1]
        idx = 0

    # The final hash (message to sign) is the hash of:
    # 1) hash of the prefix ||
    # 2) hash of the witness for signing ||
    # 3) the hash type (encoded as a 4-byte little-endian value)
    buf = BytesIO()
    buf.write(struct.pack(b'<I', hashType))

    # Optimization for SIGHASH_ALL. In this case, the prefix hash is
    # the same as the transaction hash because only the inputs have
    # been modified, so don't bother to do the wasteful O(N^2) extra
    # hash here.
    # The caching only works if the "anyone can pay flag" is also
    # disabled.
    if cachedPrefix and \
            (hashType & sigHashMask == SigHashAll) and \
            (hashType & SigHashAnyOneCanPay == 0) and \
            sig_hash_optimization:
        prefixHash = cachedPrefix
    else:
        prefixHash = txCopy.tx_sha()

    # If the ValueIn is to be included in what we're signing, sign
    # the witness hash that includes it. Otherwise, just sign the
    # prefix and signature scripts.
    if hashType & sigHashMask != SigHashAllValue:
        witnessHash = txCopy.tx_sha_witness_signing()
    else:
        witnessHash = txCopy.tx_sha_witness_value_signing()

    buf.write(prefixHash)
    buf.write(witnessHash)
    buf.seek(0) # Necessary?
    return DecredHash(buf.read())

def asSmallInt(op):
    """Returns the passed opcode as an integer."""
    if op.value == OP_0:
        return 0
    return op.value - (OP_1 - 1)

def getSigOpCount(pops, precise):
    """Counts the number of signature operations.

    If precise is True, then we attempt to count the number of operations
    for a multisig op. Otherwise we use the maximum.
    """
    nSigs = 0
    for i, pop in enumerate(pops):
        if pop.opcode.value in [OP_CHECKSIG, OP_CHECKSIGVERIFY, OP_CHECKSIGALT, OP_CHECKSIGALTVERIFY,
                                OP_CHECKMULTISIG, OP_CHECKMULTISIGVERIFY]:
            # If precise, look for familiar patterns for multisig.
            if precise and i > 0 and \
                    pops[i-1].opcode.value >= OP_1 and \
                    pops[i-1].opcode.value <= OP_16:
                nSigs += asSmallInt(pops[i-1].opcode)
            else:
                nSigs += MaxPubKeysPerMultiSig

    return nSigs

def GetSigOpCount(script):
    """Count the signature operations in a script.

    CHECKSIG counts as 1. CHECK_MULTISIG counts as 20.
    """
    pops = parseScript(script)
    return getSigOpCount(pops, False)

def GetPreciseSigOpCount(scriptSig, scriptPubKey, bip16):
    """Returns the number of signature operations in scriptPubKey.

    If bip16 is True, then scriptSig may be searched for the P2SH script
    in order to find the precise number of signature operations.
    """
    pops = parseScript(scriptPubKey)

    # Treat non-P2SH as normal.
    if not (bip16 and isScriptHash(pops)):
        return getSigOpCount(pops, True)

    # The public key script is P2SH. Parse the signature script.
    # Scripts that fail to parse count as 0 signature operations.
    try:
        sigPops = parseScript(scriptSig)
    except Exception:
        return 0
    # The signature script must only push data to the stack.
    if not isPushOnly(sigPops) or len(sigPops) == 0:
        return 0

    # P2SH script is the last item the signature script pushes.
    shScript = sigPops[-1].data
    if len(shScript) == 0:
        return 0

    # Parse the P2SH script.
    shPops = parseScript(shScript)
    return getSigOpCount(shPops, True)




import unittest

from decred.core.script import scriptnum

hexToBytes = lambda s: s.decode('hex')

class ScriptnumTest(unittest.TestCase):
    def test_number_to_bytearray(self):
        number_tests = (
            (1, hexToBytes("01")),
            (-1, hexToBytes("81")),
            (127, hexToBytes("7f")),
            (-127, hexToBytes("ff")),
            (128, hexToBytes("8000")),
            (-128, hexToBytes("8080")),
            (129, hexToBytes("8100")),
            (-129, hexToBytes("8180")),
            (256, hexToBytes("0001")),
            (-256, hexToBytes("0081")),
            (32767, hexToBytes("ff7f")),
            (-32767, hexToBytes("ffff")),
            (32768, hexToBytes("008000")),
            (-32768, hexToBytes("008080")),
            (65535, hexToBytes("ffff00")),
            (-65535, hexToBytes("ffff80")),
            (524288, hexToBytes("000008")),
            (-524288, hexToBytes("000088")),
            (7340032, hexToBytes("000070")),
            (-7340032, hexToBytes("0000f0")),

            # Values that are out of range for data that is interpreted as
            # numbers, but are allowed as the result of numeric operations.
            (2147483648, hexToBytes("0000008000")),
            (-2147483648, hexToBytes("0000008080")),
            (2415919104, hexToBytes("0000009000")),
            (-2415919104, hexToBytes("0000009080")),
            (4294967295, hexToBytes("ffffffff00")),
            (-4294967295, hexToBytes("ffffffff80")),
            (4294967296, hexToBytes("0000000001")),
            (-4294967296, hexToBytes("0000000081")),
            (281474976710655, hexToBytes("ffffffffffff00")),
            (-281474976710655, hexToBytes("ffffffffffff80")),
            (72057594037927935, hexToBytes("ffffffffffffff00")),
            (-72057594037927935, hexToBytes("ffffffffffffff80")),
            (9223372036854775807, hexToBytes("ffffffffffffff7f")),
            (-9223372036854775807, hexToBytes("ffffffffffffffff")),
        )

        for num, serialized in number_tests:
            result = scriptnum.num2bytearray(num)
            self.assertEqual(serialized, result)

    def test_create_script_num(self):
        number_tests = (
            # Minimal encoding must reject negative 0.
            (hexToBytes("80"), 0, True, scriptnum.StackMinimalDataError),

            # Minimally encoded valid values with minimal encoding flag.
            # Should not error and return expected integral number.
            (b'', 0, True, False),
            (hexToBytes("01"), 1, True, False),
            (hexToBytes("81"), -1, True, False),
            (hexToBytes("7f"), 127, True, False),
            (hexToBytes("ff"), -127, True, False),
            (hexToBytes("8000"), 128, True, False),
            (hexToBytes("8080"), -128, True, False),
            (hexToBytes("8100"), 129, True, False),
            (hexToBytes("8180"), -129, True, False),
            (hexToBytes("0001"), 256, True, False),
            (hexToBytes("0081"), -256, True, False),
            (hexToBytes("ff7f"), 32767, True, False),
            (hexToBytes("ffff"), -32767, True, False),
            (hexToBytes("008000"), 32768, True, False),
            (hexToBytes("008080"), -32768, True, False),
            (hexToBytes("ffff00"), 65535, True, False),
            (hexToBytes("ffff80"), -65535, True, False),
            (hexToBytes("000008"), 524288, True, False),
            (hexToBytes("000088"), -524288, True, False),
            (hexToBytes("000070"), 7340032, True, False),
            (hexToBytes("0000f0"), -7340032, True, False),
            (hexToBytes("00008000"), 8388608, True, False),
            (hexToBytes("00008080"), -8388608, True, False),
            (hexToBytes("ffffff7f"), 2147483647, True, False),
            (hexToBytes("ffffffff"), -2147483647, True, False),


            # Minimally encoded values that are out of range for data that
            # is interpreted as script numbers with the minimal encoding
            # flag set.  Should error and return 0.
            (hexToBytes("0000008000"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("0000008080"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("0000009000"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("0000009080"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffff00"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffff80"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("0000000001"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("0000000081"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffff00"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffff80"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffffff00"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffffff80"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffffff7f"), 0, True, scriptnum.StackNumberTooBigError),
            (hexToBytes("ffffffffffffffff"), 0, True, scriptnum.StackNumberTooBigError),


            # Non-minimally encoded, but otherwise valid values with
            # minimal encoding flag.  Should error and return 0.
            (hexToBytes("00"), 0, True, scriptnum.StackMinimalDataError),       # 0
            (hexToBytes("0100"), 0, True, scriptnum.StackMinimalDataError),     # 1
            (hexToBytes("7f00"), 0, True, scriptnum.StackMinimalDataError),     # 127
            (hexToBytes("800000"), 0, True, scriptnum.StackMinimalDataError),   # 128
            (hexToBytes("810000"), 0, True, scriptnum.StackMinimalDataError),   # 129
            (hexToBytes("000100"), 0, True, scriptnum.StackMinimalDataError),   # 256
            (hexToBytes("ff7f00"), 0, True, scriptnum.StackMinimalDataError),   # 32767
            (hexToBytes("00800000"), 0, True, scriptnum.StackMinimalDataError), # 32768
            (hexToBytes("ffff0000"), 0, True, scriptnum.StackMinimalDataError), # 65535
            (hexToBytes("00000800"), 0, True, scriptnum.StackMinimalDataError), # 524288
            (hexToBytes("00007000"), 0, True, scriptnum.StackMinimalDataError), # 7340032

            # Non-minimally encoded, but otherwise valid values without
            # minimal encoding flag.  Should not error and return expected
            # integral number.
            (hexToBytes("00"), 0, False, False),
            (hexToBytes("0100"), 1, False, False),
            (hexToBytes("7f00"), 127, False, False),
            (hexToBytes("800000"), 128, False, False),
            (hexToBytes("810000"), 129, False, False),
            (hexToBytes("000100"), 256, False, False),
            (hexToBytes("ff7f00"), 32767, False, False),
            (hexToBytes("00800000"), 32768, False, False),
            (hexToBytes("ffff0000"), 65535, False, False),
            (hexToBytes("00000800"), 524288, False, False),
            (hexToBytes("00007000"), 7340032, False, False),
        )

        for serialized, num, minimal_encoding, exception_type in number_tests:
            if exception_type:
                self.assertRaises(exception_type, scriptnum.script_num, serialized,
                                  minimal_encoding, scriptnum.math_op_code_max_script_num_len)
            else:
                scr_num = scriptnum.script_num(serialized, minimal_encoding, 
                                scriptnum.math_op_code_max_script_num_len)
                self.assertEqual(num, scr_num)


    def test_int_32(self):
        number_tests = (
            # Values inside the valid int32 range are just the values
            # themselves cast to an int32.
            (0, 0),
            (1, 1),
            (-1, -1),
            (127, 127),
            (-127, -127),
            (128, 128),
            (-128, -128),
            (129, 129),
            (-129, -129),
            (256, 256),
            (-256, -256),
            (32767, 32767),
            (-32767, -32767),
            (32768, 32768),
            (-32768, -32768),
            (65535, 65535),
            (-65535, -65535),
            (524288, 524288),
            (-524288, -524288),
            (7340032, 7340032),
            (-7340032, -7340032),
            (8388608, 8388608),
            (-8388608, -8388608),
            (2147483647, 2147483647),
            (-2147483647, -2147483647),
            (-2147483648, -2147483648),

            # Values outside of the valid int32 range are limited to int32.
            (2147483648, 2147483647),
            (-2147483649, -2147483648),
            (1152921504606846975, 2147483647),
            (-1152921504606846975, -2147483648),
            (2305843009213693951, 2147483647),
            (-2305843009213693951, -2147483648),
            (4611686018427387903, 2147483647),
            (-4611686018427387903, -2147483648),
            (9223372036854775807, 2147483647),
            (-9223372036854775808, -2147483648),
        )

        for num, expected in number_tests:
            result = scriptnum.int_32(num)
            self.assertEqual(expected, result)



from __future__ import absolute_import
import struct

from bitcoin.core import b2x, b2lx
from bitcoin.core.serialize import ser_read, Serializable, VectorSerializer

from .serialize import DecredHash
from .transaction import Transaction

class BlockHeader(Serializable):
    """A Decred block header."""
    __slots__ = ['version', 'prev_block', 'merkle_root', 'stake_root',
                 'vote_bits', 'final_state', 'voters', 'fresh_stake',
                 'revocations', 'pool_size', 'bits', 'sbits', 'height',
                 'size', 'timestamp', 'nonce', 'extra_data']

    def __init__(self, version=1, prev_block=b'\x00'*32, merkle_root=b'\x00'*32,
                 stake_root=b'\x00'*32, vote_bits=0, final_state=b'\x00'*6, voters=0,
                 fresh_stake=0, revocations=0, pool_size=0, bits=0, sbits=0,
                 height=0, size=0, timestamp=0, nonce=0, extra_data=b'\x00'*36):
        if not len(prev_block) == 32:
            raise ValueError('Previous block hash must be 32 bytes.')
        if not len(merkle_root) == 32:
            raise ValueError('Merkle root must be 32 bytes.')
        if not len(stake_root) == 32:
            raise ValueError('Stake root must be 32 bytes.')
        if not len(extra_data) == 36:
            raise ValueError('Extra data must be 32 bytes.')
        args = (version, prev_block, merkle_root, stake_root, vote_bits, final_state,
                voters, fresh_stake, revocations, pool_size, bits, sbits, height,
                size, timestamp, nonce, extra_data,)
        for attr_name, argument in zip(BlockHeader.__slots__, args):
            object.__setattr__(self, attr_name, argument)

    def GetHash(self):
        return DecredHash(self.serialize())

    @classmethod
    def stream_deserialize(cls, f):
        version = struct.unpack(b'<i', ser_read(f, 4))[0]
        prev_block = ser_read(f, 32)
        merkle_root = ser_read(f, 32)
        stake_root = ser_read(f, 32)
        vote_bits = struct.unpack(b'<H', ser_read(f, 2))[0]
        final_state = ser_read(f, 6)
        voters = struct.unpack(b'<H', ser_read(f, 2))[0]
        fresh_stake = struct.unpack(b'<B', ser_read(f, 1))[0]
        revocations = struct.unpack(b'<B', ser_read(f, 1))[0]
        pool_size = struct.unpack(b'<I', ser_read(f, 4))[0]
        bits = struct.unpack(b'<I', ser_read(f, 4))[0]
        sbits = struct.unpack(b'<q', ser_read(f, 8))[0]
        height = struct.unpack(b'<I', ser_read(f, 4))[0]
        size = struct.unpack(b'<I', ser_read(f, 4))[0]
        timestamp = struct.unpack(b'<I', ser_read(f, 4))[0]
        nonce = struct.unpack(b'<I', ser_read(f, 4))[0]
        extra_data = ser_read(f, 36)
        return cls(version, prev_block, merkle_root, stake_root, vote_bits, final_state,
                   voters, fresh_stake, revocations, pool_size, bits, sbits, height,
                   size, timestamp, nonce, extra_data)

    def stream_serialize(self, f):
        f.write(struct.pack(b'<i', self.version))
        assert len(self.prev_block) == 32
        f.write(self.prev_block)
        assert len(self.merkle_root) == 32
        f.write(self.merkle_root)
        assert len(self.stake_root) == 32
        f.write(self.stake_root)
        f.write(struct.pack(b'<H', self.vote_bits))
        assert len(self.final_state) == 6
        f.write(self.final_state)
        f.write(struct.pack(b'<H', self.voters))
        f.write(struct.pack(b'<B', self.fresh_stake))
        f.write(struct.pack(b'<B', self.revocations))
        f.write(struct.pack(b'<I', self.pool_size))
        f.write(struct.pack(b'<I', self.bits))
        f.write(struct.pack(b'<q', self.sbits))
        f.write(struct.pack(b'<I', self.height))
        f.write(struct.pack(b'<I', self.size))
        f.write(struct.pack(b'<I', self.timestamp))
        f.write(struct.pack(b'<I', self.nonce))
        assert len(self.extra_data) == 36
        f.write(self.extra_data)

class Block(BlockHeader):
    """A Decred block."""
    __slots__ = ['txs', 'stxs']


    def __init__(self, version=1, prev_block=b'\x00'*32, merkle_root=b'\x00'*32,
                 stake_root=b'\x00'*32, vote_bits=0, final_state=b'\x00'*6, voters=0,
                 fresh_stake=0, revocations=0, pool_size=0, bits=0, sbits=0,
                 height=0, size=0, timestamp=0, nonce=0, extra_data=b'\x00'*36,
                 txs=(), stxs=()):

        args = (version, prev_block, merkle_root, stake_root, vote_bits, final_state,
                voters, fresh_stake, revocations, pool_size, bits, sbits, height,
                size, timestamp, nonce, extra_data,)
        super(Block, self).__init__(*args)
        self.txs = list(txs)
        self.stxs = list(stxs)

    @classmethod
    def stream_deserialize(cls, f):
        self = super(Block, cls).stream_deserialize(f)
        txs = VectorSerializer.stream_deserialize(Transaction, f)
        stxs = VectorSerializer.stream_deserialize(Transaction, f)

        self.txs = list(txs)
        self.stxs = list(stxs)
        return self

    def stream_serialize(self, f):
        super(Block, self).stream_serialize(f)
        VectorSerializer.stream_serialize(Transaction, self.txs, f)
        VectorSerializer.stream_serialize(Transaction, self.stxs, f)

    def get_header(self):
        """Return the block header as a new object."""
        args = ['version', 'prev_block', 'merkle_root', 'stake_root', 'vote_bits',
                'final_state', 'voters', 'fresh_stake', 'revocations', 'pool_size',
                'bits', 'sbits', 'height', 'size', 'timestamp', 'nonce', 'extra_data']
        kwargs = {}
        for i in args:
            kwargs[i] = getattr(self, i)
        return BlockHeader(**kwargs)

    def GetHash(self):
        return super(Block, self).GetHash()



from __future__ import absolute_import

from .script import *
from .engine import *
from .opcode import *
from .scriptbuilder import *
from .constants import *

### TODO ###
# This file is only partially ported from dcrd. #
############

# Max bytes allowed in pushed data for a nulldate transaction.
maxDataCarrierSize = 256

StandardVerifyFlags = ScriptBip16 | \
        ScriptVerifyDERSignatures | \
        ScriptVerifyStrictEncoding | \
        ScriptVerifyMinimalData | \
        ScriptDiscourageUpgradableNops | \
        ScriptVerifyCleanStack | \
        ScriptVerifyCheckLockTimeVerify | \
        ScriptVerifyLowS


# Classes of script payment known about in the blockchain.
NonStandardTy       = 0             # None of the recognized forms.
PubKeyTy            = 1             # Pay pubkey.
PubKeyHashTy        = 2             # Pay pubkey hash.
ScriptHashTy        = 3             # Pay to script hash.
MultiSigTy          = 4             # Multi signature.
NullDataTy          = 5             # Empty data-only (provably prunable).
StakeSubmissionTy   = 6             # Stake submission.
StakeGenTy          = 7             # Stake generation
StakeRevocationTy   = 8             # Stake revocation.
StakeSubChangeTy    = 9             # Change for stake submission tx.
PubkeyAltTy         = 10            # Alternative signature pubkey.
PubkeyHashAltTy     = 11            # Alternative signature pubkey hash.

scriptClassToName = {
    NonStandardTy:     "nonstandard",
    PubKeyTy:          "pubkey",
    PubkeyAltTy:       "pubkeyalt",
    PubKeyHashTy:      "pubkeyhash",
    PubkeyHashAltTy:   "pubkeyhashalt",
    ScriptHashTy:      "scripthash",
    MultiSigTy:        "multisig",
    NullDataTy:        "nulldata",
    StakeSubmissionTy: "stakesubmission",
    StakeGenTy:        "stakegen",
    StakeRevocationTy: "stakerevoke",
    StakeSubChangeTy:  "sstxchange",
}

# Moved here from script.py
def HasP2SHScriptSigStakeOpCodes(version, scriptSig, scriptPubKey):
    _class = GetScriptClass(version, scriptPubKey)
    if IsStakeOutput(scriptPubKey):
        _class = GetStakeOutSubclass(scriptPubKey)
    if _class == ScriptHashTy:
        # Obtain the embedded pkScript from scriptSig.
        # Ensure it does not use and stake tagging opcodes.
        shScript = GetPkScriptFromP2SHSigScript(scriptSig)
        hasStakeOpCodes = ContainsStakeOpCodes(shScript)
        if hasStakeOpCodes:
            raise P2SHStakeOpCodesError()

def isPubkey(pops):
    """Returns true if the script passed is a pay-to-pubkey transaction."""
    return (len(pops) == 2 and
        (len(pops[0].data) == 33 or len(pops[0].data) == 65) and
        pops[1].opcode.value == OP_CHECKSIG)

def isOneByteMaxDataPush(po):
    """Returns true if the parsed opcode pushed exactly one byte to the stack."""
    return po.opcode.value in [OP_1, OP_2, OP_3, OP_4, OP_5, OP_6,
            OP_7, OP_8, OP_9, OP_10, OP_11, OP_12, OP_13, OP_14,
            OP_15, OP_16, OP_DATA_1]

def isPubkeyAlt(pops):
    """Returns true if the script is an alternative pay-to-pubkey transaction."""
    return (len(pops) == 3 and
        len(pops[0].data) < 512 and
        isOneByteMaxDataPush(pops[1]) and
        pops[2].opcode.value == OP_CHECKSIGALT)

def isPubkeyHash(pops):
    """Returns true if the script is a pay-to-pubkey-hash transaction."""
    return (len(pops) == 5 and
        pops[0].opcode.value == OP_DUP and
        pops[1].opcode.value == OP_HASH160 and
        pops[2].opcode.value == OP_DATA_20 and
        pops[3].opcode.value == OP_EQUALVERIFY and
        pops[4].opcode.value == OP_CHECKSIG)

def isPubkeyHashAlt(pops):
    """Returns true if the script is a pay-to-pubkey-hash transaction."""
    return (len(pops) == 6 and
        pops[0].opcode.value == OP_DUP and
        pops[1].opcode.value == OP_HASH160 and
        pops[2].opcode.value == OP_DATA_20 and
        pops[3].opcode.value == OP_EQUALVERIFY and
        isOneByteMaxDataPush(pops[4]) and
        pops[5].opcode.value == OP_CHECKSIGALT)

def isScriptHash(pops):
    """Returns true is the script is a pay-to-script-hash transaction."""
    return (len(pops) == 3 and
        pops[0].opcode.value == OP_HASH160 and
        pops[1].opcode.value == OP_DATA_20 and
        pops[2].opcode.value == OP_EQUAL)

def isAnyKindOfScriptHash(pops):
    """Returns true if the script is a pay-to-script-hash or stake pay-to-script-hash transaction."""
    standardP2SH = (len(pops) == 3 and
        pops[0].opcode.value == OP_HASH160 and
        pops[1].opcode.value == OP_DATA_20 and
        pops[2].opcode.value == OP_EQUAL)
    if standardP2SH:
        return True

    stakeP2SH = (len(pops) == 4 and
        (pops[0].opcode.value >= 186 and pops[0].opcode.value <= 189) and
        pops[1].opcode.value == OP_HASH160 and
        pops[2].opcode.value == OP_DATA_20 and
        pops[3].opcode.value == OP_EQUAL)
    if stakeP2SH:
        return True
    return False

def isMultiSig(pops):
    """Returns true if the script is a multisig transaction."""
    # Absolute minimum is 1 pubkey.
    if len(pops) < 4:
        return False
    if not isSmallInt(pops[0].opcode):
        return False
    if not isSmallInt(pops[-2].opcode):
        return False
    if pops[-1].opcode.value != OP_CHECKMULTISIG:
        return False

    # Verify the number of pubkeys specified matches the actual number provided.
    if len(pops) - 3 != asSmallInt(pops[-2].opcode):
        return False

    for pop in pops[1:-2]:
        if len(pop.data) != 33 and len(pop.data) != 65:
            return False
    return True

def isMultisigScript(script):
    """Returns whether script is a multisignature script."""
    pops = parseScript(script)
    return isMultiSig(pops)

def isMultisigSigScript(script):
    """Returns whether script is a multisignature script."""
    if len(script) == 0 or script is None:
        return False
    pops = parseScript(script)
    subPops = parseScript(pops[-1].data)
    return isMultiSig(subPops)

def isNullData(pops):
    """Returns whether script is a null data transaction."""
    if len(pops) == 1 and pops[0].opcode.value == OP_RETURN:
        return True

    return (len(pops) == 2 and
        pops[0].opcode.value == OP_RETURN and
        pops[1].opcode.value <= OP_PUSHDATA4 and
        len(pops[1].data) <= maxDataCarrierSize)

def isStakeSubmission(pops):
    """Returns whether script is a stake submission transaction."""
    if (len(pops) == 6 and
            pops[0].opcode.value == OP_SSTX and
            pops[1].opcode.value == OP_DUP and
            pops[2].opcode.value == OP_HASH160 and
            pops[3].opcode.value == OP_DATA_20 and
            pops[4].opcode.value == OP_EQUALVERIFY and
            pops[5].opcode.value == OP_CHECKSIG):
        return True

    if (len(pops) == 4 and
            pops[0].opcode.value == OP_SSTX and
            pops[1].opcode.value == OP_HASH160 and
            pops[2].opcode.value == OP_DATA_20 and
            pops[3].opcode.value == OP_EQUAL):
        return True

    return False

def isStakeGen(pops):
    """Returns whether the script is a stake generation transaction."""
    if (len(pops) == 6 and
            pops[0].opcode.value == OP_SSGEN and
            pops[1].opcode.value == OP_DUP and
            pops[2].opcode.value == OP_HASH160 and
            pops[3].opcode.value == OP_DATA_20 and
            pops[4].opcode.value == OP_EQUALVERIFY and
            pops[5].opcode.value == OP_CHECKSIG):
        return True

    if (len(pops) == 4 and
            pops[0].opcode.value == OP_SSGEN and
            pops[1].opcode.value == OP_HASH160 and
            pops[2].opcode.value == OP_DATA_20 and
            pops[3].opcode.value == OP_EQUAL):
        return True

    return False

def isStakeRevocation(pops):
    """Returns whether the script is a stake submission revocation transaction."""
    if (len(pops) == 6 and
            pops[0].opcode.value == OP_SSRTX and
            pops[1].opcode.value == OP_DUP and
            pops[2].opcode.value == OP_HASH160 and
            pops[3].opcode.value == OP_DATA_20 and
            pops[4].opcode.value == OP_EQUALVERIFY and
            pops[5].opcode.value == OP_CHECKSIG):
        return True

    if (len(pops) == 4 and
            pops[0].opcode.value == OP_SSRTX and
            pops[1].opcode.value == OP_HASH160 and
            pops[2].opcode.value == OP_DATA_20 and
            pops[3].opcode.value == OP_EQUAL):
        return True

    return False

def isSStxChange(pops):
    """Returns whether script is a stake submission change transaction."""
    if (len(pops) == 6 and
            pops[0].opcode.value == OP_SSTXCHANGE and
            pops[1].opcode.value == OP_DUP and
            pops[2].opcode.value == OP_HASH160 and
            pops[3].opcode.value == OP_DATA_20 and
            pops[4].opcode.value == OP_EQUALVERIFY and
            pops[5].opcode.value == OP_CHECKSIG):
        return True

    if (len(pops) == 4 and
            pops[0].opcode.value == OP_SSTXCHANGE and
            pops[1].opcode.value == OP_HASH160 and
            pops[2].opcode.value == OP_DATA_20 and
            pops[3].opcode.value == OP_EQUAL):
        return True

    return False

def typeOfScript(pops):
    """Returns the type of the script being inspected from the known standard types."""
    if isPubkey(pops):
        return PubKeyTx
    elif isPubkeyAlt(pops):
        return PubkeyAltTy
    elif isPubkeyHashAlt(pops):
        return PubkeyHashAlt
    elif isScriptHash(pops):
        return ScriptHashTy
    elif isMultiSig(pops):
        return MultiSigTy
    elif isNullData(pops):
        return NullDataTy
    elif isStakeSubmission(pops):
        return StakeSubmissionTy
    elif isStakeGen(pops):
        return StakeGenTy
    elif isStakeRevocation(pops):
        return StakeRevocationTy
    elif isSStxChange(pops):
        return StakeSubChangeTy

    return NonStandardTy

def GetScriptClass(version, script):
    """Returns the class of the script.

    If the script does not parse, NonStandardTy is returned.
    """
    # NullDataTy outputs can have non-default script versions.
    if version != DefaultScriptVersion:
        return NonStandardTy

    try:
        pops = parseScript(script)
    except Exception:
        return NonStandardTy
    else:
        return typeOfScript(pops)

def expectedInputs(pops, _class, subclass):
    """Returns the number of arguments required by a script.

    If the script is of unknown type such that the number can not be determined
    then -1 is returned. We are an internal function and thus assume that class
    is the real class of pops (and we can thus assume things that were determined
    while finding out the type).
    """
    if _class == PubKeyTy:
        return 1
    elif _class == PubKeyHashTy:
        return 2
    elif _class == StakeSubmissionTy:
        if subclass == PubKeyHashTy:
            return 2
        return 1 # P2SH
    elif _class == StakeGenTy:
        if subclass == PubKeyHashTy:
            return 2
        return 1 # P2SH
    elif _class == StakeRevocationTy:
        if subclass == PubKeyHashTy:
            return 2
        return 1 # P2SH
    elif _class == StakeSubChangeTy:
        if subclass == PubKeyHashTy:
            return 2
        return 1 # P2SH
    elif _class == ScriptHashTy:
        # Not including script, handled below.
        return 1
    elif _class == MultiSigTy:
        # Standard multisig pushes a small int for the number of sigs
        # and number of keys.
        return asSmallInt(pops[0].opcode)

    elif _class == NullDataTy:
        pass

    return -1

class ScriptInfo(object):
    def __init__(self, PkScriptClass=0, NumInputs=0, ExpectedInputs=0, SigOps=0):
        self.PkScriptClass = PkScriptClass
        self.NumInputs = NumInputs
        self.ExpectedInputs = ExpectedInputs
        self.SigOps = SigOps

def IsStakeOutput(pkScript):
    """Returns whether an output script is a stake type."""
    pkPops = parseScript(pkScript)

    _class = typeOfScript(pkPops)
    return _class in [StakeSubmissionTy, StakeGenTy, StakeRevocationTy, StakeSubChangeTy]

def GetStakeOutSubclass(pkScript):
    """Extracts the subclass (P2PKH or P2SH) from a stake output."""
    pkPops = parseScript(pkScript)

    _class = typeOfScript(pkPops)
    isStake = _class in [StakeSubmissionTy, StakeGenTy, StakeRevocationTy, StakeSubChangeTy]
    if isStake:
        stakeSubscript = []
        for pop in pkPops:
            if pop.opcode.value >= 186 and pop.opcode.value <= 189:
                continue
            stakeSubscript.append(pop)
        subClass = typeOfScript(stakeSubscript)
    else:
        return 0

    return subClass

def getStakeOutSubscript(pkScript):
    """Extracts the subscript (P2PKH or P2SH) from a stake output."""
    return pkScript[1:]

def GetPkScriptFromP2SHSigScript(sigScript):
    """Returns the embedded pkScript from the signature script of a P2SH-spending tx."""
    sigPops = parseScript(sigScript)
    # The pay-to-hash-script is the final data push.
    return sigPops[-1].data

def ContainsStakeOpCodes(pkScript):
    """Returns whether a pkScript contains stake tagging opcodes."""
    shPops = parseScript(pkScript)
    for pop in shPops:
        if pop.opcode.value >= 186 and pop.opcode.value <= 189:
            return True

    return False

def CalcScriptInfo(sigScript, pkScript, bip16):
    """Returns a structure providing data about the provided script pair."""
    sigPops = parseScript(sigScript)
    pkPops = parseScript(pkScript)

    pkScriptClass = typeOfScript(pkPops)
    if not isPushOnly(sigPops):
        raise StackNonPushOnlyError()

    subClass = ScriptClass(0)
    if pkScriptClass in [StakeSubmissionTy, StakeGenTy, StakeRevocationTy, StakeSubChangeTy]:
        subClass = GetStakeOutSubclass(pkScript)

    ExpectedInputs = expectedInputs(pkPops, pkScriptClass, subClass)
    NumInputs = len(sigPops)

    if pkScriptClass == ScriptHashTy and bip16:
        # P2SH is the final data push.
        script = sigPops[-1].data
        shPops = parseScript(script)

        shInputs = expectedInputs(shPops, typeOfScript(shPops), 0)
        if shInputs == -1:
            ExpectedInputs = -1
        else:
            ExpectedInputs += shInputs
        SigOps = getSigOpCount(shPops, True)
    else:
        SigOps = getSigOpCount(pkPops, True)

    return ScriptInfo(pkScriptClass, NumInputs, ExpectedInputs, SigOps)

def CalcMultiSigStats(script):
    """Returns the number of pubkeys and sigs from a multisig script.

    The script MUST already be known to be a multisig script.
    """
    pops = parseScript(script)
    if len(pops) < 4:
        raise StackUnderflowError()

    numSigs = asSmallInt(pops[0].opcode)
    numPubKeys = asSmallInt(pops[-2].opcode)
    return numPubKeys, numSigs

def MultisigRedeemScriptFromScriptSig(script):
    """Extracts a redeem script from a P2SH-redeeming input."""
    pops = parseScript(script)
    return pops[-1].data

def payToPubKeyHashScript(pubKeyHash):
    """Creates a new output script to a 20-byte pubkey hash."""
    builder = ScriptBuilder()
    return builder.AddOp(OP_DUP).AddOp(OP_HASH160).AddData(pubKeyHash).AddOp(OP_EQUALVERIFY).AddOp(OP_CHECKSIG).Script()

def payToPubKeyHashEdwardsScript(pubKeyHash):
    """Creates a new output script to a 20-byte Edwards pubkey hash."""
    edwardsData = _bchr(edwards)
    builder = ScriptBuilder()
    return builder.AddOp(OP_DUP).AddOp(OP_HASH160).AddData(pubKeyHash).AddOp(OP_EQUALVERIFY).AddData(edwardsData).AddOp(OP_CHECKSIGALT).Script()

def payToPubKeyHashSchnorrScript(pubKeyHash):
    """Creates a new output script to a 20-byte pubkey hash.

    Expects a schnorr signature instead of a secp256k1 signature.
    """
    schnorrData = _bchr(secSchnorr)
    builder = ScriptBuilder()
    return builder.AddOp(OP_DUP).AddOp(OP_HASH160).AddData(pubKeyHash).AddOp(OP_EQUALVERIFY).AddData(schnorrData).AddOp(OP_CHECKSIGALT).Script()

def payToScriptHash(scriptHash):
    """Creates a new output script to pay to a script hash."""
    builder = ScriptBuilder()
    return builder.AddOp(OP_HASH160).AddData(scriptHash).AddOp(OP_EQUAL).Script()

def GetScriptHashFromP2SHScript(pkScript):
    """Extracts the script hash from a valid P2SH pkScript."""
    pops = parseScript(pkScript)
    reachedHash160DataPush = False

    sh = b''
    for p in pops:
        if p.opcode.value == OP_HASH160:
            reachedHash160DataPush = True
            continue
        if reachedHash160DataPush:
            sh = p.data

    return sh

def PayToScriptHashScript(scriptHash):
    return payToScriptHashScript(scriptHash)

def payToPubKeyScript(serializedPubKey):
    """Creates a new output script to pay to a public key."""
    builder = ScriptBuilder()
    return builder.AddData(serializedPubKey).AddOp(OP_CHECKSIG).Script()

def payToEdwardsPubKeyScript(serializedPubKey):
    """Creates a new output script to pay to an Ed25519 public key."""
    builder = ScriptBuilder()
    edwardsData = _bchr(edwards)
    return builder.AddData(serializedPubKey).AddData(edwardsData).AddOp(OP_CHECKSIGALT).Script()

def payToSchnorrPubKeyScript(serializedPubKey):
    """Creates a new output script to pay to a public key, but to be signed with a Schnorr signature."""
    builder = ScriptBuilder()
    schnorrData = _bchr(secSchnorr)
    return builder.AddData(serializedPubKey).AddData(schnorrData).AddOp(OP_CHECKSIGALT).Script()

def PayToSStx(addr):
    """Creates a script to pay to a script hash or pubkey hash, but tags the output with OP_SSTX."""
    raise Exception('Not implemented')



from __future__ import absolute_import
import sys
import struct

from .errors import *
from .opcode import *
from .scriptnum import num2bytearray
from .constants import maxScriptSize, MaxScriptElementSize

_bchr = chr
_bord = ord
if sys.version > '3':
    _bchr = lambda x: bytes([x])
    _bord = lambda x: x

# defaultScriptAlloc is the default size used for the backing array
# for a script being built by the ScriptBuilder.  The array will
# dynamically grow as needed, but this figure is intended to provide
# enough space for vast majority of scripts without needing to grow the
# backing array multiple times.
defaultScriptAlloc = 500

# ErrScriptNotCanonical identifies a non-canonical script.  The caller can use
# a type assertion to detect this error type.
##type ErrScriptNotCanonical string

#// Error implements the error interface.
#func (e ErrScriptNotCanonical) Error() string {
#	return string(e)
#}

class ScriptNonCanonicalError(Exception):
    pass

# ScriptBuilder provides a facility for building custom scripts.  It allows
# you to push opcodes, ints, and data while respecting canonical encoding.  In
# general it does not ensure the script will execute correctly, however any
# data pushes which would exceed the maximum allowed script engine limits and
# are therefore guaranteed not to execute will not be pushed and will result in
# the Script function returning an error.
#
# For example, the following would build a 2-of-3 multisig script for usage in
# a pay-to-script-hash (although in this situation MultiSigScript() would be a
# better choice to generate the script):
# 	builder := txscript.NewScriptBuilder()
# 	builder.AddOp(txscript.OP_2).AddData(pubKey1).AddData(pubKey2)
# 	builder.AddData(pubKey3).AddOp(txscript.OP_3)
# 	builder.AddOp(txscript.OP_CHECKMULTISIG)
# 	script, err := builder.Script()
# 	if err != nil {
# 		// Handle the error.
# 		return
# 	}
# 	fmt.Printf("Final multi-sig script: %x\n", script)
class ScriptBuilder(object):
    def __init__(self):
        self.script = []
        self.err = None

    def AddOp(self, opcode):
        """AddOp pushes the passed opcode to the end of the script.

        The script will not be modified if pushing the opcode would cause
        the script to exceed the maximum allowed script engine size.
        """
        if self.err:
            return self

        if len(self._makeScript()) + 1 > maxScriptSize:
            self.err = ScriptNonCanonicalError('adding an opcode would exceed the maximum allowed canonical script length')
            return self

        self.script.append(_bchr(opcode))
        return self

    def canonicalDataSize(self, data):
        """Returns the number of bytes the canonical encoding of data will take."""
        dl = len(data)

	# When the data consists of a single number that can be represented
	# by one of the "small integer" opcodes, that opcode will be instead
	# of a data push opcode followed by the number.
        if dl == 0:
            return 1
        elif dl == 1 and _bord(data[0]) <= 16:
            return 1
        elif dl == 1 and _bord(data[0]) == 0x81:
            return 1

        if dl < OP_PUSHDATA1:
            return 1 + dl
        elif dl <= 0xff:
            return 2 + dl
        elif dl <= 0xffff:
            return 3 + dl

        return 5 + dl

    def _addData(self, data):
        dl = len(data)

	# When the data consists of a single number that can be represented
	# by one of the "small integer" opcodes, use that opcode instead of
	# a data push opcode followed by the number.
        if dl == 0 or dl == 1 and _bord(data[0]) == 0:
            self.script.append(_bchr(OP_0))
            return self
        elif dl == 1 and _bord(data[0]) <= 16:
            self.script.append(_bchr((OP_1-1)+_bord(data[0])))
            return self
        elif dl == 1 and _bord(data[0]) == 0x81:
            self.script.append(_bchr(OP_1NEGATE))
            return self

	# Use one of the OP_DATA_# opcodes if the length of the data is small
	# enough so the data push instruction is only a single byte.
	# Otherwise, choose the smallest possible OP_PUSHDATA# opcode that
	# can represent the length of the data.
        if dl < OP_PUSHDATA1:
            self.script.append(_bchr((OP_DATA_1-1)+dl))
        elif dl <= 0xff:
            self.script.append(_bchr(OP_PUSHDATA1))
            self.script.append(_bchr(dl))
        elif dl <= 0xffff:
            buf = struct.pack(b'<H', dl)
            self.script.append(_bchr(OP_PUSHDATA2))
            self.script.append(buf)
        else:
            buf = struct.pack(b'<I', dl)
            self.script.append(_bchr(OP_PUSHDATA4))
            self.script.append(buf)

        # Append the actual data.
        self.script.append(data)
        return self

    def AddFullData(self, data):
        """AddFullData should not typically be used by ordinary users as it does not
        include the checks which prevent data pushes larger than the maximum allowed
        sizes which leads to scripts that can't be executed.  This is provided for
        testing purposes such as regression tests where sizes are intentionally made
        larger than allowed.

        Use AddData instead.
        """
        if self.err:
            return self

        return self._addData(data)

    def AddData(self, data):
        """AddData pushes the passed data to the end of the script.
        
        It automatically chooses canonical opcodes depending on the length of the data.
        A zero length buffer will lead to a push of empty data onto the stack (OP_0) and any push
        of data greater than MaxScriptElementSize will not modify the script since
        that is not allowed by the script engine.  Also, the script will not be
        modified if pushing the data would cause the script to exceed the maximum
        allowed script engine size.
        """
        if self.err:
            return self

        # Pushes that would cause the script to exceed the largest allowed
        # script size would result in a non-canonical script.
        dsize = self.canonicalDataSize(data)
        if len(self._makeScript()) + dsize > maxScriptSize:
            self.err = ScriptNonCanonicalError('adding %d bytes of data would exceed the maximum allowed canonical script length' % dsize)
            return self

        # Pushes larger than the max script element size would result in a
        # script that is not canonical.
        dl = len(data)
        if dl > MaxScriptElementSize:
            self.err = ScriptNonCanonicalError('adding a data element of %d bytes would exceed the maximum allowed canonical script length' % dl)
            return self

        return self._addData(data)

    def AddInt64(self, val):
        """Pushes the passed integer to the end of the script.

        The script will not be modified if pushing the data would cause
        the script to exceed the maximum allowed script engine size.
        """
        if self.err:
            return self

	# Pushes that would cause the script to exceed the largest allowed
	# script size would result in a non-canonical script.
        if len(self._makeScript()) + 1 > maxScriptSize:
            self.err = ScriptNonCanonicalError('adding an integer would exceed the maximum allowed canonical script length')
            return self

        # Fast path for small integers and OP_1NEGATE
        if val == 0:
            self.script.append(_bchr(OP_0))
            return self
        if val == -1 or (val >= 1 and val <= 16):
            self.script.append(_bchr((OP_1-1)+val))
            return self

        return self.AddData(num2bytearray(val))

    def Reset(self):
        """Resets the script so it has no contents."""
        self.script = self.script[0:0]
        self.err = None
        return self

    def _makeScript(self):
        return ''.join(self.script)

    def Script(self, raise_on_error=False):
        """Returns the currently built script.

        When any errors occurred while building the script, the script will
        be returned up to the point of the first error.
        """
        if raise_on_error and self.err:
            raise self.err
        return self._makeScript(), self.err

    def get_script(self):
        s, _ = self.Script(raise_on_error=True)
        return s



import unittest
import struct

from bitcoin.core import x, b2x, b2lx

from decred.core.transaction import *

class TransactionTest(unittest.TestCase):

    def test_transaction_deserialization_and_serialization(self):
        # Transaction from Decred unit test.
        raw_tx = x('01000000010000000000000000000000000000000000000000000000000000000000000000ffffffff00ffffffff0200f2052a01000000abab434104d64bdfd09eb1c5fe295abdeb1dca4281be988e2da0b6c1c6a59dc226c28624e18175e851c96b973d81b01cc31f047834bc06d6d6edf620d184241a6aed8b63a6ac00e1f50500000000bcbc434104d64bdfd09eb1c5fe295abdeb1dca4281be988e2da0b6c1c6a59dc226c28624e18175e851c96b973d81b01cc31f047834bc06d6d6edf620d184241a6aed8b63a6ac00000000000000000112121212121212121515151534343434070431dc001b0162')
        tx = Transaction.deserialize(raw_tx)

        self.assertEqual(1, tx.version)
        self.assertEqual(0, tx.locktime)
        self.assertEqual(0, tx.expiry)

        # Test input.
        self.assertEqual(1, len(tx.txins))
        txin = tx.txins[0]
        self.assertEqual(1302123111085380114, txin.value)
        self.assertEqual(353703189, txin.block_height)
        self.assertEqual(875836468, txin.block_index)
        self.assertEqual(x('0431dc001b0162'), txin.sig_script)
        self.assertEqual(4294967295, txin.sequence)

        self.assertEqual(2, len(tx.txouts))
        # Test output 0.
        txout = tx.txouts[0]
        self.assertEqual(5000000000, txout.value)
        self.assertEqual(43947, txout.version)
        self.assertEqual(x('4104d64bdfd09eb1c5fe295abdeb1dca4281be988e2da0b6c1c6a59dc226c28624e18175e851c96b973d81b01cc31f047834bc06d6d6edf620d184241a6aed8b63a6ac'), txout.pk_script)
        # Test output 1.
        txout = tx.txouts[1]
        self.assertEqual(100000000, txout.value)
        self.assertEqual(48316, txout.version)
        self.assertEqual(x('4104d64bdfd09eb1c5fe295abdeb1dca4281be988e2da0b6c1c6a59dc226c28624e18175e851c96b973d81b01cc31f047834bc06d6d6edf620d184241a6aed8b63a6ac'), txout.pk_script)

        self.assertEqual(b2x(raw_tx), b2x(tx.serialize()))

    def test_vars_to_tx_version(self):
        def repr_version(int32):
            return b2x(struct.pack(b'<i', int32))

        self.assertEqual('01000000', repr_version(vars_to_tx_version(1, TX_SERIALIZE_FULL)))
        self.assertEqual('01000100', repr_version(vars_to_tx_version(1, TX_SERIALIZE_NO_WITNESS)))



import unittest
import sys

from decred.core.script import opcode, scriptbuilder
from decred.core.script.constants import maxScriptSize
from decred.core.script.errors import *
from decred.core.script.scriptnum import num2bytearray

_bchr = chr
if sys.version > '3':
    _bchr = lambda x: bytes([x])

# ByteArray
BA = lambda items: b''.join(map(_bchr, items))

class ScriptBuilderTest(unittest.TestCase):
    def test_add_op(self):
        tests = (
            ('push OP_0', [opcode.OP_0], b'\x00'),
            ('push OP_1 OP_2', [opcode.OP_1, opcode.OP_2], b'QR'),
            ('push OP_HASH160 OP_EQUAL', [opcode.OP_HASH160, opcode.OP_EQUAL], b'\xa9\x87'),
        )

        builder = scriptbuilder.ScriptBuilder()
        for name, opcodes, expected in tests:
            builder.Reset()
            for op in opcodes:
                builder.AddOp(op)
            result, err = builder.Script()
            self.assertIs(None, err)
            self.assertEqual(expected, result)

    def test_add_int_64(self):
        tests = (
            ("push -1", -1, BA([opcode.OP_1NEGATE])),
            ("push small int 0", 0, BA([opcode.OP_0])),
            ("push small int 1", 1, BA([opcode.OP_1])),
            ("push small int 2", 2, BA([opcode.OP_2])),
            ("push small int 3", 3, BA([opcode.OP_3])),
            ("push small int 4", 4, BA([opcode.OP_4])),
            ("push small int 5", 5, BA([opcode.OP_5])),
            ("push small int 6", 6, BA([opcode.OP_6])),
            ("push small int 7", 7, BA([opcode.OP_7])),
            ("push small int 8", 8, BA([opcode.OP_8])),
            ("push small int 9", 9, BA([opcode.OP_9])),
            ("push small int 10", 10, BA([opcode.OP_10])),
            ("push small int 11", 11, BA([opcode.OP_11])),
            ("push small int 12", 12, BA([opcode.OP_12])),
            ("push small int 13", 13, BA([opcode.OP_13])),
            ("push small int 14", 14, BA([opcode.OP_14])),
            ("push small int 15", 15, BA([opcode.OP_15])),
            ("push small int 16", 16, BA([opcode.OP_16])),
            ("push 17", 17, BA([opcode.OP_DATA_1, 0x11])),
            ("push 65", 65, BA([opcode.OP_DATA_1, 0x41])),
            ("push 127", 127, BA([opcode.OP_DATA_1, 0x7f])),
            ("push 128", 128, BA([opcode.OP_DATA_2, 0x80, 0])),
            ("push 255", 255, BA([opcode.OP_DATA_2, 0xff, 0])),
            ("push 256", 256, BA([opcode.OP_DATA_2, 0, 0x01])),
            ("push 32767", 32767, BA([opcode.OP_DATA_2, 0xff, 0x7f])),
            ("push 32768", 32768, BA([opcode.OP_DATA_3, 0, 0x80, 0])),
            ("push -2", -2, BA([opcode.OP_DATA_1, 0x82])),
            ("push -3", -3, BA([opcode.OP_DATA_1, 0x83])),
            ("push -4", -4, BA([opcode.OP_DATA_1, 0x84])),
            ("push -5", -5, BA([opcode.OP_DATA_1, 0x85])),
            ("push -17", -17, BA([opcode.OP_DATA_1, 0x91])),
            ("push -65", -65, BA([opcode.OP_DATA_1, 0xc1])),
            ("push -127", -127, BA([opcode.OP_DATA_1, 0xff])),
            ("push -128", -128, BA([opcode.OP_DATA_2, 0x80, 0x80])),
            ("push -255", -255, BA([opcode.OP_DATA_2, 0xff, 0x80])),
            ("push -256", -256, BA([opcode.OP_DATA_2, 0x00, 0x81])),
            ("push -32767", -32767, BA([opcode.OP_DATA_2, 0xff, 0xff])),
            ("push -32768", -32768, BA([opcode.OP_DATA_3, 0x00, 0x80, 0x80])),
        )

        builder = scriptbuilder.ScriptBuilder()
        for name, val, expected in tests:
            builder.Reset().AddInt64(val)
            result, err = builder.Script()
            self.assertIs(None, err)
            self.assertEqual(expected, result)

    def test_add_data(self):
        tests = (

            # BIP0062: Pushing an empty byte sequence must use OP_0.
            ("push empty byte sequence", b'', BA([opcode.OP_0]), False),
            ("push 1 byte 0x00", BA([0x00]), BA([opcode.OP_0]), False),

            # BIP0062: Pushing a 1-byte sequence of byte 0x01 through 0x10 must use OP_n.
            ("push 1 byte 0x01", BA([0x01]), BA([opcode.OP_1]), False),
            ("push 1 byte 0x02", BA([0x02]), BA([opcode.OP_2]), False),
            ("push 1 byte 0x03", BA([0x03]), BA([opcode.OP_3]), False),
            ("push 1 byte 0x04", BA([0x04]), BA([opcode.OP_4]), False),
            ("push 1 byte 0x05", BA([0x05]), BA([opcode.OP_5]), False),
            ("push 1 byte 0x06", BA([0x06]), BA([opcode.OP_6]), False),
            ("push 1 byte 0x07", BA([0x07]), BA([opcode.OP_7]), False),
            ("push 1 byte 0x08", BA([0x08]), BA([opcode.OP_8]), False),
            ("push 1 byte 0x09", BA([0x09]), BA([opcode.OP_9]), False),
            ("push 1 byte 0x0a", BA([0x0a]), BA([opcode.OP_10]), False),
            ("push 1 byte 0x0b", BA([0x0b]), BA([opcode.OP_11]), False),
            ("push 1 byte 0x0c", BA([0x0c]), BA([opcode.OP_12]), False),
            ("push 1 byte 0x0d", BA([0x0d]), BA([opcode.OP_13]), False),
            ("push 1 byte 0x0e", BA([0x0e]), BA([opcode.OP_14]), False),
            ("push 1 byte 0x0f", BA([0x0f]), BA([opcode.OP_15]), False),
            ("push 1 byte 0x10", BA([0x10]), BA([opcode.OP_16]), False),

            # BIP0062: Pushing the byte 0x81 must use OP_1NEGATE.
            ("push 1 byte 0x81", BA([0x81]), BA([opcode.OP_1NEGATE]), False),

            # BIP0062: Pushing any other byte sequence up to 75 bytes must
            # use the normal data push (opcode byte n, with n the number of
            # bytes, followed n bytes of data being pushed).
            ("push 1 byte 0x11", BA([0x11]), BA([opcode.OP_DATA_1, 0x11]), False),
            ("push 1 byte 0x80", BA([0x80]), BA([opcode.OP_DATA_1, 0x80]), False),
            ("push 1 byte 0x82", BA([0x82]), BA([opcode.OP_DATA_1, 0x82]), False),
            ("push 1 byte 0xff", BA([0xff]), BA([opcode.OP_DATA_1, 0xff]), False),
            (
                    "push data len 17",
                    BA([0x49] * 17),
                    BA([opcode.OP_DATA_17]) + BA([0x49] * 17),
                    False
            ),
            (
                    "push data len 75",
                    BA([0x49] * 75),
                    BA([opcode.OP_DATA_75]) + BA([0x49] * 75),
                    False
            ),
            # BIP0062: Pushing 76 to 255 bytes must use OP_PUSHDATA1.
            (
                    "push data len 76",
                    BA([0x49] * 76),
                    BA([opcode.OP_PUSHDATA1, 76]) + BA([0x49] * 76),
                    False
            ),
            (
                    "push data len 255",
                    BA([0x49] * 255),
                    BA([opcode.OP_PUSHDATA1, 255]) + BA([0x49] * 255),
                    False
            ),

            # BIP0062: Pushing 256 to 520 bytes must use OP_PUSHDATA2.
            (
                    "push data len 256",
                    BA([0x49] * 256),
                    BA([opcode.OP_PUSHDATA2, 0, 1]) + BA([0x49] * 256),
                    False
            ),
            (
                    "push data len 520",
                    BA([0x49] * 520),
                    BA([opcode.OP_PUSHDATA2, 0x08, 0x02]) + BA([0x49] * 520),
                    False
            ),

            # BIP0062: OP_PUSHDATA4 can never be used, as pushes over 520
            # bytes are not allowed, and those below can be done using
            # other operators.
            (
                    "push data len 521",
                    BA([0x49] * 4097),
                    b'',
                    False
            ),
            (
                    "push data len 32767 (canonical)",
                    BA([0x49] * 32767),
                    b'',
                    False
            ),
            (
                    "push data len 65536 (canonical)",
                    BA([0x49] * 65536),
                    b'',
                    False
            ),
            # Additional tests for the PushFullData function that
            # intentionally allows data pushes to exceed the limit for
            # regression testing purposes.

            # 3-byte data push via OP_PUSHDATA_2.
            (
                    "push data len 32767 (non-canonical)",
                    BA([0x49] * 32767),
                    BA([opcode.OP_PUSHDATA2, 255, 127]) + BA([0x49] * 32767),
                    True
            ),

            # 5-byte data push via OP_PUSHDATA_4.
            (
                    "push data len 65536 (non-canonical)",
                    BA([0x49] * 65536),
                    BA([opcode.OP_PUSHDATA4, 0, 0, 1, 0]) + BA([0x49] * 65536),
                    True
            ),

        )

        builder = scriptbuilder.ScriptBuilder()
        for name, data, expected, use_full in tests:
            builder.Reset()
            if not use_full:
                builder.AddData(data)
            else:
                builder.AddFullData(data)

            if expected == b'':
                self.assertRaises(Exception, builder.Script, True)
            else:
                result, err = builder.Script()
                self.assertIs(None, err)
                self.assertEqual(expected, result)

    def test_exceed_max_script_size(self):
        # Construct a max-size script.
        builder = scriptbuilder.ScriptBuilder()
        builder.Reset().AddFullData(b'\x00' * (maxScriptSize - 3))
        orig_script, err = builder.Script()
        self.assertIs(None, err)

        # Ensure adding data that would exceed maximum size does not work.
        script, err = builder.AddData(b'\x00').Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding an opcode that would exceed maximum size does not work.
        builder.Reset().AddFullData(b'\x00' * (maxScriptSize - 3))
        script, err = builder.AddOp(opcode.OP_0).Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding an integer that would exceed maximum size does not work.
        builder.Reset().AddFullData(b'\x00' * (maxScriptSize - 3))
        script, err = builder.AddInt64(0).Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

    def test_errored_script(self):
        builder = scriptbuilder.ScriptBuilder()
        builder.Reset().AddFullData(b'\x00' * (maxScriptSize - 8))
        orig_script, err = builder.Script()
        self.assertIs(None, err)

        script, err = builder.AddData(b'\x00' * 5).Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding data using the non-canonical path to a script that has errored doesn't succeed.
        script, err = builder.AddFullData(b'\x00').Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding data to a script that has errored doesn't succeed.
        script, err = builder.AddData(b'\x00').Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding an opcode to a script that has errored doesn't succeed.
        script, err = builder.AddOp(opcode.OP_0).Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure adding an integer to a script that has errored doesn't succeed.
        script, err = builder.AddInt64(0).Script()
        self.assertIsInstance(err, scriptbuilder.ScriptNonCanonicalError)
        self.assertEqual(script, orig_script)

        # Ensure the error has a message set.
        self.assertNotEqual('', err.message)



from __future__ import absolute_import
import sys

from bitcoin.core import _bignum

from .errors import *

_bord = ord
if sys.version > '3':
    _bord = lambda x: x

max_int_32 = (1 << 31) - 1
min_int_32 = -1 << 31

# Max number of bytes data being interpreted as an int may be
# for the majority of opcodes.
math_op_code_max_script_num_len = 4
# Max number of bytes for the type of alternative signature suite.
alt_sig_suites_max_script_num_len = 1

def check_minimal_data_encoding(val):
    """Returns whether or not the passed bytearray adheres to the
    minimal encoding requirements."""
    if len(val) == 0:
        return

    # If the most-significant byte (excluding the sign bit) is zero,
    # it's not minimal.
    if _bord(val[-1]) & 0x7f == 0:
        # One exception in which there's more than one byte and the most
        # significant bit of the second-most significant byte is set it would
        # conflict with the sign bit.
        if len(val) == 1 or _bord(val[len(val)-2]) & 0x80 == 0:
            raise StackMinimalDataError()

def int_32(val):
    """Returns the script number clamped to valid int32 values."""
    val = min(val, max_int_32)
    val = max(val, min_int_32)
    return val

def num2bytearray(n):
    """Convert an integer to encoded bytes."""
    return _bignum.bn2vch(n)

def script_num(v, require_minimal=False, script_num_len=math_op_code_max_script_num_len):
    """Convert encoded bytes to an integer."""
    if len(v) > script_num_len:
        raise StackNumberTooBigError()

    if require_minimal:
        check_minimal_data_encoding(v)

    result = _bignum.vch2bn(v)
    return result




from __future__ import absolute_import
import struct
from io import BytesIO

from bitcoin.core import b2lx, b2x
from bitcoin.core.serialize import (ser_read, Serializable, BytesSerializer,
        VectorSerializer, VarIntSerializer)

TX_SERIALIZE_FULL = 0
TX_SERIALIZE_NO_WITNESS = 1
TX_SERIALIZE_ONLY_WITNESS = 2
TX_SERIALIZE_WITNESS_SIGNING = 3
TX_SERIALIZE_WITNESS_VALUE_SIGNING = 4

def tx_version_to_vars(version):
    """Converts an int32 representation into serialization type and version."""
    f = BytesIO()
    f.write(struct.pack(b'<I', version))
    f.seek(0)
    version_value = struct.unpack(b'<H', ser_read(f, 2))[0]
    serialization_type = struct.unpack(b'<H', ser_read(f, 2))[0]
    return (version_value, serialization_type)

def vars_to_tx_version(version_value, serialization_type):
    """Converts serialization type and tx version into a 32 bit int."""
    f = BytesIO()
    f.write(struct.pack(b'<H', version_value))
    f.write(struct.pack(b'<H', serialization_type))
    f.seek(0)
    version = struct.unpack(b'<i', ser_read(f, 4))[0]
    return version

class OutPoint(Serializable):
    """A Decred previous transaction output."""
    __slots__ = ['hash', 'index', 'tree']

    def __init__(self, hash=b'\x00'*32, index=0, tree=0):
        if not len(hash) == 32:
            raise ValueError('Previous hash must be 32 bytes.')
        object.__setattr__(self, 'hash', hash)
        object.__setattr__(self, 'index', index)
        object.__setattr__(self, 'tree', tree)

    def __repr__(self):
        return 'OutPoint(hash=lx(%r), index=%d, tree=%d)' % (b2lx(self.hash), self.index, self.tree)

    @classmethod
    def from_outpoint(cls, outpoint):
        return cls(outpoint.hash, outpoint.index, outpoint.tree)

    @classmethod
    def stream_deserialize(cls, f):
        hash = ser_read(f, 32)
        index = struct.unpack(b'<I', ser_read(f, 4))[0]
        tree = struct.unpack(b'<b', ser_read(f, 1))[0]
        return cls(hash, index, tree)

    def stream_serialize(self, f):
        assert len(self.hash) == 32
        f.write(self.hash)
        f.write(struct.pack(b'<I', self.index))
        f.write(struct.pack(b'<b', self.tree))

class TxIn(Serializable):
    """A Decred transaction input."""
    __slots__ = ['prev_out', 'sequence', 'value', 'block_height', 'block_index', 'sig_script']

    def __init__(self, prev_out=OutPoint(), sequence=0, value=0, block_height=0, block_index=0, sig_script=b''):
        # Non-witness.
        object.__setattr__(self, 'prev_out', prev_out)
        object.__setattr__(self, 'sequence', sequence)
        # Witness.
        object.__setattr__(self, 'value', value)
        object.__setattr__(self, 'block_height', block_height)
        object.__setattr__(self, 'block_index', block_index)
        object.__setattr__(self, 'sig_script', sig_script)

    def __repr__(self):
        return 'TxIn(prev_out=%s, sequence=0x%x, value=%d, block_height=%d, block_index=%d, sig_script=x("%s"))' % (repr(self.prev_out),
                self.sequence, self.value, self.block_height, self.block_index, b2x(self.sig_script))

    @classmethod
    def from_txin(cls, txin):
        return cls(OutPoint.from_outpoint(txin.prev_out), txin.sequence, txin.value, txin.block_height, txin.block_index, txin.sig_script)

    @classmethod
    def stream_deserialize(cls, f):
        prev_out = OutPoint.stream_deserialize(f)
        sequence = struct.unpack(b'<I', ser_read(f, 4))[0]

        value = struct.unpack(b'<q', ser_read(f, 8))[0]
        block_height = struct.unpack(b'<I', ser_read(f, 4))[0]
        block_index = struct.unpack(b'<I', ser_read(f, 4))[0]
        sig_script = BytesSerializer.stream_deserialize(f)
        return cls(prev_out, sequence, value, block_height, block_index, sig_script)

    def deserialize_prefix(self, f):
        prev_out = OutPoint.stream_deserialize(f)
        sequence = struct.unpack(b'<I', ser_read(f, 4))[0]

        self.prev_out = prev_out
        self.sequence = sequence

    def deserialize_witness(self, f):
        value = struct.unpack(b'<q', ser_read(f, 8))[0]
        block_height = struct.unpack(b'<I', ser_read(f, 4))[0]
        block_index = struct.unpack(b'<I', ser_read(f, 4))[0]
        sig_script = BytesSerializer.stream_deserialize(f)

        self.value = value
        self.block_height = block_height
        self.block_index = block_index
        self.sig_script = sig_script

    def deserialize_witness_signing(self, f):
        sig_script = BytesSerializer.stream_deserialize(f)
        self.sig_script = sig_script

    def deserialize_witness_value_signing(self, f):
        value = struct.unpack(b'<q', ser_read(f, 8))[0]
        sig_script = BytesSerializer.stream_deserialize(f)

        self.value = value
        self.sig_script = sig_script

    def stream_serialize(self, f):
        OutPoint.stream_serialize(self.prev_out, f)
        f.write(struct.pack(b'<I', self.sequence))

        f.write(struct.pack(b'<q', self.value))
        f.write(struct.pack(b'<I', self.block_height))
        f.write(struct.pack(b'<I', self.block_index))
        BytesSerializer.stream_serialize(self.sig_script, f)

    def serialize_prefix(self, f):
        OutPoint.stream_serialize(self.prev_out, f)
        f.write(struct.pack(b'<I', self.sequence))

    def serialize_witness(self, f):
        f.write(struct.pack(b'<q', self.value))
        f.write(struct.pack(b'<I', self.block_height))
        f.write(struct.pack(b'<I', self.block_index))
        BytesSerializer.stream_serialize(self.sig_script, f)

    def serialize_witness_signing(self, f):
        BytesSerializer.stream_serialize(self.sig_script, f)

    def serialize_witness_value_signing(self, f):
        f.write(struct.pack(b'<q', self.value))
        BytesSerializer.stream_serialize(self.sig_script, f)

class TxOut(Serializable):
    """A Decred transaction output."""
    __slots__ = ['value', 'version', 'pk_script']

    def __init__(self, value=0, version=0, pk_script=b'\x00'):
        object.__setattr__(self, 'value', value)
        object.__setattr__(self, 'version', version)
        object.__setattr__(self, 'pk_script', pk_script)

    def __repr__(self):
        return 'TxOut(value=%d, version=%d, pk_script=x("%s"))' % (self.value, self.version, b2x(self.pk_script))

    @classmethod
    def from_txout(cls, txout):
        return cls(txout.value, txout.version, txout.pk_script)

    @classmethod
    def stream_deserialize(cls, f):
        value = struct.unpack(b'<q', ser_read(f, 8))[0]
        version = struct.unpack(b'<H', ser_read(f, 2))[0]
        pk_script = BytesSerializer.stream_deserialize(f)
        return cls(value, version, pk_script)

    def stream_serialize(self, f):
        f.write(struct.pack(b'<q', self.value))
        f.write(struct.pack(b'<H', self.version))
        BytesSerializer.stream_serialize(self.pk_script, f)

class Transaction(Serializable):
    """A decred transaction."""
    __slots__ = ['version', 'txins', 'txouts', 'locktime', 'expiry']

    def __init__(self, version=0, txins=(), txouts=(), locktime=0, expiry=0):
        object.__setattr__(self, 'version', version)
        object.__setattr__(self, 'txins', list(TxIn.from_txin(txin) for txin in txins))
        object.__setattr__(self, 'txouts', list(TxOut.from_txout(txout) for txout in txouts))
        object.__setattr__(self, 'locktime', locktime)
        object.__setattr__(self, 'expiry', expiry)

    def __repr__(self):
        return 'Transaction(version=%d, txins=%r, txouts=%r, locktime=%d, expiry=%d)' % (self.version,
                self.txins, self.txouts, self.locktime, self.expiry)

    @classmethod
    def from_tx(cls, tx):
        return cls(tx.version, tx.txins, tx.txouts, tx.locktime, tx.expiry)

    @classmethod
    def stream_deserialize(cls, f):
        self = cls()
        self.version = struct.unpack(b'<i', ser_read(f, 4))[0]
        _, ser_type = tx_version_to_vars(self.version)

        if ser_type == TX_SERIALIZE_NO_WITNESS:
            self.deserialize_prefix(f)
        elif ser_type == TX_SERIALIZE_ONLY_WITNESS:
            self.deserialize_witness(f, False)
        elif ser_type == TX_SERIALIZE_WITNESS_SIGNING:
            self.deserialize_witness_signing(f)
        elif ser_type == TX_SERIALIZE_WITNESS_VALUE_SIGNING:
            self.deserialize_witness_value_signing(f)
        elif ser_type == TX_SERIALIZE_FULL:
            self.deserialize_prefix(f)
            self.deserialize_witness(f, True)
        else:
            raise Exception('Unsupported transaction type.')

        return self

    def deserialize_prefix(self, f):
        """Deserialize the transaction prefix."""
        txin_count = VarIntSerializer.stream_deserialize(f)
        txins = []
        for i in range(txin_count):
            txin = TxIn()
            txin.deserialize_prefix(f)
            txins.append(txin)

        txouts = VectorSerializer.stream_deserialize(TxOut, f)
        locktime = struct.unpack(b'<I', ser_read(f, 4))[0]
        expiry = struct.unpack(b'<I', ser_read(f, 4))[0]

        self.txins = list(txins)
        self.txouts = list(txouts)
        self.locktime = locktime
        self.expiry = expiry

    def deserialize_witness(self, f, txins_present):
        """Deserialize the transaction witnesses.

        If txins_present is True, existing txins will be populated
        with witness data.
        """
        txin_count = VarIntSerializer.stream_deserialize(f)
        if not txins_present:
            txins = []
            for i in range(txin_count):
                txin = TxIn()
                txin.deserialize_witness(f)
                txins.append(txin)
            self.txins = list(txins)
        else:
            for i in range(txin_count):
                txin = self.txins[i]
                txin.deserialize_witness(f)

    def deserialize_witness_signing(self, f):
        """Deserialize a witness for signing."""
        txin_count = VarIntSerializer.stream_deserialize(f)
        txins = []
        for i in range(txin_count):
            txin = TxIn()
            txin.deserialize_witness_signing(f)
            txins.append(txin)
        self.txins = list(txins)

    def deserialize_witness_value_signing(self, f):
        """Deserialize a witness for signing with value."""
        txin_count = VarIntSerializer.stream_deserialize(f)
        txins = []
        for i in range(txin_count):
            txin = TxIn()
            txin.deserialize_witness_value_signing(f)
            txins.append(txin)
        self.txins = list(txins)

    def stream_serialize(self, f):
        f.write(struct.pack(b'<i', self.version))
        _, ser_type = tx_version_to_vars(self.version)

        if ser_type == TX_SERIALIZE_NO_WITNESS:
            self.serialize_prefix(f)
        elif ser_type == TX_SERIALIZE_ONLY_WITNESS:
            self.serialize_witness(f)
        elif ser_type == TX_SERIALIZE_WITNESS_SIGNING:
            self.serialize_witness_signing(f)
        elif ser_type == TX_SERIALIZE_WITNESS_VALUE_SIGNING:
            self.serialize_witness_value_signing(f)
        elif ser_type == TX_SERIALIZE_FULL:
            self.serialize_prefix(f)
            self.serialize_witness(f)
        else:
            raise Exception('Unsupported transaction type.')

    def serialize_prefix(self, f):
        VarIntSerializer.stream_serialize(len(self.txins), f)
        for txin in self.txins:
            txin.serialize_prefix(f)

        VectorSerializer.stream_serialize(TxOut, self.txouts, f)
        f.write(struct.pack(b'<I', self.locktime))
        f.write(struct.pack(b'<I', self.expiry))

    def serialize_witness(self, f):
        VarIntSerializer.stream_serialize(len(self.txins), f)
        for txin in self.txins:
            txin.serialize_witness(f)

    def serialize_witness_signing(self, f):
        VarIntSerializer.stream_serialize(len(self.txins), f)
        for txin in self.txins:
            txin.serialize_witness_signing(f)

    def serialize_witness_value_signing(self, f):
        VarIntSerializer.stream_serialize(len(self.txins), f)
        for txin in self.txins:
            txin.serialize_witness_value_signing(f)




#!/usr/bin/env python

from setuptools import setup, find_packages
import os

version = '0.1a'
requires = ['decred_hash', 'python-bitcoinlib']

setup(
    name='pydecred',
    author='Tyler Willis',
    author_email='kefkiusrex@gmail.com',
    version=version,
    url='https://github.com/kefkius/pydecred',
    description='Decred python library.',
    keywords='decred',
    packages=find_packages(),
    install_requires=requires,
    test_suite='decred.tests'
)



import blake_hash
import hashlib

def DecredHash(msg):
    return blake_hash.getPoWHash(msg)

def Hash160(msg):
    """RIPEMD_160(BLAKE256(msg))"""
    h = hashlib.new('ripemd160')
    h.update(DecredHash(msg))
    return h.digest()
